"""You can ignore this module and any files within it. It is used to help set up your User Data Functions.
"""

# flake8: noqa: F401
from .docstring_parser import extract_summary_description
from .spec_utils import get_multipart_content


"""This module contains the classes that are used to connect to resources like Warehouses, SQL databases, 
and Lakehouses. You can use these classes by importing them from the fabric.functions module 
(e.g. `from fabric.functions import FabricSqlConnection`).
"""

import base64
import json
import time
import requests
import struct
import typing
import logging

# flake8: noqa: I005
import pyodbc
from azure.storage.filedatalake import DataLakeDirectoryClient

from fabric.functions.fabric_item import FabricItem
from fabric.internal.fabric_lakehouse_files_client import FabricLakehouseFilesClient
from fabric.functions.udf_exception import UserDataFunctionInternalError
from azure.storage.filedatalake.aio import DataLakeDirectoryClient as DataLakeDirectoryClientAsync

class FabricSqlConnection(FabricItem):
    """
    This class is used to connect to resources that supply SQL connection strings.

    A User Data Function with a parameter of this type must be decorated with :meth:`fabric.functions.UserDataFunctions.connection` (see the example under `Remarks`).


    .. remarks::

        .. note::
            If you want to connect to a Lakehouse SQL endpoint, that should be done through
            :class:`FabricLakehouseClient` instead (and then use it's method :meth:`FabricLakehouseClient.connectToSql`).
        

        To use this class and have Fabric make the proper connections to your SQL endpoint, you must:
        * Add a data connection in the Connections tab of your User Data Functions on the portal.
        * Add a parameter to your User Data Function with the type 'FabricSqlConnection'.
        * Add the decorator `connection` to your User Data Function that references the parameter
        and the alias of the data connection you made.
        
        .. code-block:: python
                import fabric.functions as fn

                udf = fn.UserDataFunctions()
    
                @udf.connection("<data connection alias>", "<argName>")
                @udf.function()
                def my_function(<argName>: fn.FabricSqlConnection) -> None:
                    conn = <argName>.connect()
                    pass

    """

    __APPSETTINGS_PATH = "sqlendpoint"
    __INITIAL_CATALOG = "Initial Catalog"

    def _get_split_connection_string(self) -> typing.Dict[str, str]:
        connString = self.endpoints[self.__APPSETTINGS_PATH]["ConnectionString"]

        # Lakehouse connection string contains Data Source instead of Server
        connString = connString.replace("Data Source", "Server")

        if "=" not in connString:
            return {"Server": connString}

        # The reason we can't use a simple split on semi-colons is because a Fabric SQL database name can contain semi-colons.
        result = {}
        in_quotes = False
        quote_char = ""
        current_pair = []
        current_segment = []

        for char in connString:
            if char == '"' or char == "'":
                if in_quotes and char == quote_char:
                    in_quotes = False
                    quote_char = ""
                    current_segment.append(char)
                elif in_quotes:
                    current_segment.append(char)
                elif not in_quotes:
                    in_quotes = True
                    quote_char = char
                    current_segment.append(char)

            elif char == ';' and not in_quotes:
                current_segment = ''.join(current_segment).strip()
                if current_segment:
                    current_pair.append(current_segment)
                if len(current_pair) == 2:
                    key, value = current_pair

                    result[key] = value
                current_pair = []
                current_segment = []
            elif char == '=' and not in_quotes and not current_pair:
                current_segment = ''.join(current_segment).strip()
                if current_segment:
                    current_pair.append(current_segment)
                current_segment = []
            else:
                current_segment.append(char)

        # Handle the last pair
        current_segment = ''.join(current_segment).strip()
        if current_segment:
            current_pair.append(current_segment)
        if len(current_pair) == 2:
            key, value = current_pair

            result[key] = value

        if "Server" not in result:
            raise UserDataFunctionInternalError("Server value is not set in connection")

        return result

    def _modify_database_name(self, database_string: str) -> str:
        if '{' in database_string or '}' in database_string:
            raise UserDataFunctionInternalError("Databases connected to User Data Functions cannot contain '{' or '}' characters in their name.")

        # Remove surrounding quotes if present as they can interfere with the ODBC connection.
        # The database name should be passed without quote characters to the connection string, and then wrapped in curly braces.

        if database_string.startswith("\"") and database_string.endswith("\""):
            database_string = database_string[1:-1]
            # Sql databases try to escape quotes by doubling them inside double quoted strings, so we need to remove the extra quotes.
            if '""' in database_string:
                database_string = database_string.replace('""', '"')
        elif database_string.startswith("'") and database_string.endswith("'"):
            database_string = database_string[1:-1]   

        return f"Database={{{database_string}}};"

    def connect(self) -> pyodbc.Connection:
        """Connects to the SQL endpoint. Call this within your function before attempting to query the endpoint.
        """
        if self.__APPSETTINGS_PATH not in self.endpoints:
            raise UserDataFunctionInternalError(f"{self.__APPSETTINGS_PATH} is not set")
        
        dict_conn_string = self._get_split_connection_string()
        connString = dict_conn_string["Server"]

        # https://github.com/AzureAD/azure-activedirectory-library-for-python/wiki/Connect-to-Azure-SQL-Database
        
        token = self.endpoints[self.__APPSETTINGS_PATH]["AccessToken"].encode('UTF-8')
        exptoken = b""
        for i in token:
            exptoken+=bytes({i})
            exptoken+=bytes(1)
        tokenstruct = struct.pack("=i", len(exptoken)) + exptoken

        driver_names = [x for x in pyodbc.drivers() if x.endswith(' for SQL Server')]
        latest_driver = driver_names[-1] if driver_names else None

        if latest_driver is None:
            raise UserDataFunctionInternalError("No ODBC Driver found for SQL Server. Please download the latest for your OS.")

        conn_string = f'DRIVER={{{latest_driver}}};Server={connString};Encrypt=yes;LongAsMax=yes;'

        if self.__INITIAL_CATALOG in dict_conn_string:
            databaseString = self._modify_database_name(dict_conn_string[self.__INITIAL_CATALOG])     
            conn_string += databaseString
                  
        return pyodbc.connect(conn_string, attrs_before = {1256:tokenstruct}, timeout=60)

    
class FabricLakehouseClient(FabricItem):
    """This class is used to connect to Lakehouse resources. 
    You can use this class to connect to both SQL and Files endpoints.

    A User Data Function with a parameter of this type must be decorated with :meth:`fabric.functions.UserDataFunctions.connection` (see the example under `Remarks`).

    .. remarks::
        To use this class and have Fabric make the proper connections to your Lakehouse, you must:
        * Add a data connection in the Connections tab of your User Data Functions on the portal.
        * Add a parameter to your User Data Function with the type 'FabricLakehouseClient'.
        * Add the decorator `connection` to your User Data Function that references the parameter
        and the alias of the data connection you made.
        
        .. code-block:: python
                import fabric.functions as fn

                udf = fn.UserDataFunctions()
    
                @udf.connection("<Lakehouse alias>", "<argName>")
                @udf.function()
                def my_function(<argName>: fn.FabricLakehouseClient) -> None:
                    sql_endpoint = <argName>.connectToSql()
                    files_endpoint = <argName>.connectToFiles()
                    pass
    """

    def connectToSql(self) -> pyodbc.Connection:
        """Connects to the SQL endpoint of the Lakehouse. Call this within your function before attempting to query the endpoint.
        """

        return FabricSqlConnection(self.alias_name, self.endpoints).connect()  

    def connectToFiles(self) -> DataLakeDirectoryClient:
        """Connects to the Files endpoint of the Lakehouse. Call this within your function before attempting to query the endpoint.
        """

        return FabricLakehouseFilesClient(self.alias_name, self.endpoints).connect()
        
    def connectToFilesAsync(self) -> DataLakeDirectoryClientAsync:
        """Connects to the Files endpoint of the Lakehouse. Call this within your function before attempting to query the endpoint. Returns a client that can be used in async functions.
        """
        
        return FabricLakehouseFilesClient(self.alias_name, self.endpoints).connect_async()

class UserDataFunctionContext:
    """This class is used to be able to access certain metadata about a function invocation.

    A User Data Function with a parameter of this type must be decorated with :meth:`fabric.functions.UserDataFunctions.context` (see the example under `Remarks`).

    .. remarks::

        .. note::
            Do not use the parameter name "`context`" for this (or any) parameter in your User Data Function.

        To use this class and have Fabric generate the metadata, you must:
        * Add a parameter to your User Data Function with the type 'UserDataFunctionContext'.
        * Add the decorator `context` to your User Data Function that references the parameter.
        
        .. code-block:: python
                import fabric.functions as fn

                udf = fn.UserDataFunctions()
    
                @udf.context("<argName>")
                @udf.function()
                def my_function(<argName>: fn.UserDataFunctionContext) -> None:
                    invocation_id = <argName>.invocation_id
                    pass

    :param invocationId: The unique identifier for the function invocation.
    :type invocationId: str
    :param executingUser: Metadata about the user token used to execute the function.
    :type executingUser: typing.Dict[str, str]
    """
    def __init__(self, invocationId: str, executingUser: typing.Dict[str, str]):
        """Don't worry about using the constructor to create an instance of this class. Fabric will automatically create it for you (as long as you follow the steps within `Remarks`).
        """
        self.__invocation_id = invocationId
        self.__executing_user = executingUser

    @property
    def invocation_id(self) -> str:
        """The unique identifier for the function invocation.
        """

        return self.__invocation_id
    
    @property
    def executing_user(self) -> typing.Dict[str, str]:
        """Metadata about the user token used to execute the function.

        .. remarks::
            
                The dictionary contains the following keys:
                * `Oid`: The user's object id, which is an immutable identifier for the requestor, 
                which is the verified identity of the user or service principal. 
                This ID uniquely identifies the requestor across applications. 
                We suggest using this in tandem with TenantId to be the values to perform authorization checks.
                * `PreferredUsername`: The preferred username of the user, which can be an email address, 
                phone number, generic username, or unformatted string. This is mutable.
                * `TenantId`: The tenant id of the user, which represents the tenant that the user is signed into.

        """

        return self.__executing_user


class FabricVariablesClient(FabricItem):
    """
    This class is used to connect to Fabric Variable Library resources.

    A User Data Function with a parameter of this type must be decorated with :meth:`fabric.functions.UserDataFunctions.connection` (see the example under `Remarks`).

    .. remarks::
        To use this class and have Fabric make the proper connections to your Variable Library, you must:
        * Add a data connection in the Connections tab of your User Data Functions on the portal.
        * Add a parameter to your User Data Function with the type 'FabricVariablesClient'.
        * Add the decorator `connection` to your User Data Function that references the parameter
        and the alias of the data connection you made.
        
        .. code-block:: python
                import fabric.functions as fn

                udf = fn.UserDataFunctions()
    
                @udf.connection("<Variable Library alias>", "<argName>")
                @udf.function()
                def my_function(<argName>: fn.FabricVariablesClient) -> None:
                    variables = <argName>.getVariables()
                    pass
    """

    __APPSETTINGS_PATH = "vlendpoint"

    def getVariables(self) -> typing.Dict[str, str]:
        """Connects to the Variable Library endpoint and retrieves variables. Call this within your function before attempting to access variables.
        
        Returns:
            dict[str, str]: A dictionary mapping variable names to their values.
        """
        variables = {}
        if self.__APPSETTINGS_PATH not in self.endpoints:
            raise UserDataFunctionInternalError(f"{self.__APPSETTINGS_PATH} is not set")
            
        connection_string = self.endpoints[self.__APPSETTINGS_PATH]["ConnectionString"]
        access_token = self.endpoints[self.__APPSETTINGS_PATH]["AccessToken"]
        
        headers = {
        'Authorization': f'Bearer {access_token}',
        # 'content-type': 'application/json'
        }
    
        try:
            # Step 1 - Getting VariableLibrary info
            get_url = connection_string #f"https://api.fabric.microsoft.com/v1/workspaces/{workspace_id}/VariableLibraries/{variable_library_id}"
            
            get_response = requests.get(get_url, headers=headers)

            if get_response.status_code != 200:
                raise UserDataFunctionInternalError(f"GET request failed: {get_response.status_code} - {get_response.text}")
            
            library_info = get_response.json()
            if library_info.get('properties'):
                active_value_set_name = library_info['properties'].get('activeValueSetName')
            
            # Step 2 - Getting VariableLibrary definition from post
            post_url = f"{connection_string}/getDefinition" # f"https://api.fabric.microsoft.com/v1/workspaces/{workspace_id}/VariableLibraries/{variable_library_id}/getDefinition"            
            post_response = requests.post(post_url, headers=headers)
            
            if post_response.status_code == 202:
                status_code = 202
                location = post_response.headers.get('Location')
                retry_after = int(post_response.headers.get('Retry-After', 30))

                while status_code == 202:   
                    time.sleep(retry_after)
                    
                    polling_response = requests.get(location, headers=headers)
                    status_code = polling_response.status_code
                
                if polling_response.status_code == 200:
                    location_of_definition = polling_response.headers.get('Location')
                    definition_response = requests.get(location_of_definition, headers=headers)
                    definition_data = definition_response.json()
                else:
                    raise UserDataFunctionInternalError(f"Polling for definition failed: {polling_response.status_code} - {getattr(polling_response, 'text', '')}")
                    
            elif post_response.status_code == 200:
                definition_data = post_response.json()
            else:
                raise UserDataFunctionInternalError(f"POST request failed: {post_response.status_code} - {post_response.text}")
            
            # Step 3  - Get actual variable values
            variables = self.parse_variable_definition(definition_data, active_value_set_name)
            
        except Exception as e:
            variables = {}
            raise UserDataFunctionInternalError(f"Exception while getting variables: {str(e)}") from e
        
        return variables

    def parse_variable_definition(self, definition_data, active_value_set_name):
        """Parse variable library definition data and extract all variables and value sets"""
        try:
            definition = definition_data.get('definition', {})
            parts = definition.get('parts', [])
            default_variable_values = {}
            value_sets = {}
            for i, part in enumerate(parts):
                path = part.get('path')
                payload = part.get('payload')
                payload_type = part.get('payloadType')
                if payload_type == 'InlineBase64' and path != 'settings.json':
                    try:
                        decoded_content = base64.b64decode(payload).decode('utf-8')
                        parsed_content = json.loads(decoded_content)

                        if path == 'variables.json':
                            default_variable_values = parsed_content.get('variables', {})
                        elif path.startswith('valueSets/') and path.endswith('.json'):
                            value_set_name = path.split('/')[-1].replace('.json', '')
                            value_sets[value_set_name] = parsed_content.get('variableOverrides', {})
                    except Exception:
                        # Skip malformed parts
                        continue
            # Combine default variables with active value set overrides
            variable_values = {}
            if active_value_set_name and active_value_set_name in value_sets:
                active_values = value_sets[active_value_set_name]
                for variable in active_values:
                    # active_values is expected to be a list of dicts with 'name' and 'value'
                    if isinstance(variable, dict) and 'name' in variable and 'value' in variable:
                        variable_values[variable['name']] = variable['value']

            for variable in default_variable_values:
                if isinstance(variable, dict) and 'name' in variable and 'value' in variable:
                    if variable['name'] not in variable_values:
                        variable_values[variable['name']] = variable['value']

            return variable_values
        
            
        except Exception as e:
            return f"Parse error: {str(e)}"


# flake8: noqa : I005, I900, I001
"""This module contains the exception classes that are used to represent errors that occur during the invocation of a User Data Function.
These can be imported from the `fabric.functions` module."""

from fabric.internal.decorators.constants import (
    UDFExceptionCodes,
)


class UserDataFunctionError(Exception):
    """The exception/error base class for any errors raised during the invocation of a User Data Function.

    :param error_code: A string representing the error code associated with the exception.
    :type error_code: str
    :param message: An optional message describing the exception. Defaults to "Known User Data Function Exception Thrown".
    :type message: str
    :param properties: An optional dictionary containing additional properties related to the exception.
    :type properties: dict
    """

    def __init__(
        self,
        error_code: str,
        message: str = "Known User Data Function Exception Thrown",
        properties: dict = None,
    ):
        self.error_code = error_code
        self.message = message
        if properties is None:
            self.properties = {}
        else:
            self.properties = properties
        super().__init__(self.message)


class UserDataFunctionInternalError(UserDataFunctionError):
    """Represents an internal error that occurs during the execution of a user-defined function. This error is typically thrown when an unexpected error occurs during the execution of a function that is not related to an input or the user's code."""

    def __init__(
        self,
        message: str = "An internal execution error occured during function execution",
        properties: dict = None,
    ):
        """
        :param message: An optional message describing the exception. Defaults to "An internal execution error occured during function execution".
        :type message: str
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(UDFExceptionCodes.INTERNAL_ERROR, message, properties)


class UserDataFunctionInvalidInputError(UserDataFunctionError):
    """Represents an error that occurs when the input provided to a function is invalid. This error is typically thrown when the input provided to a function does not match the expected input type or format."""

    def __init__(
        self,
        message: str = "Something went wrong when parsing an input to this function. This could be because the provided data couldn't be constructed as the data type provided, or the provided data isn't valid JSON.",
        properties: dict = None,
    ):
        """
        :param message: An optional message describing the exception. Defaults to "Something went wrong when parsing an input to this function. This could be because the provided data couldn't be constructed as the data type provided, or the provided data isn't valid JSON.".
        :type message: str
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(UDFExceptionCodes.INVALID_INPUT, message, properties)


class UserDataFunctionMissingInputError(UserDataFunctionError):
    """Represents an error that occurs when a required parameter is missing from the input data provided to a function. This error is typically thrown when a parameter is expected to be present in the input data but is not found."""

    def __init__(
        self,
        message: str = "Parameter does not exist in binding data",
        properties: dict = None,
    ):
        """
        :param message: An optional message describing the exception. Defaults to "Parameter does not exist in binding data".
        :type message: str
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(UDFExceptionCodes.MISSING_INPUT, message, properties)


class UserDataFunctionResponseTooLargeError(UserDataFunctionError):
    """Represents an error that occurs when the response generated by a function is too large. This limit is currently set to 30 megabytes."""

    def __init__(self, limit_in_megabytes: int, properties: dict = None):
        """
        :param limit_in_megabytes: The limit in megabytes that the response exceeded.
        :type limit_in_megabytes: int
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(
            UDFExceptionCodes.RESPONSE_TOO_LARGE,
            f"Function's response size is larger than the {limit_in_megabytes} megabyte limit.",
            properties,
        )


class UserDataFunctionTimeoutError(UserDataFunctionError):
    """Represents an error that occurs when a function exceeds the timeout limit set for the function. This limit is currently set to 200 seconds."""

    def __init__(self, timeout: int, properties: dict = None):
        """
        :param timeout: The timeout limit in seconds that the function exceeded.
        :type timeout: int
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(
            UDFExceptionCodes.TIMEOUT,
            f"Function hit a timeout limit after {timeout} seconds but may still be running. Please check the logs for more information.",
            properties,
        )


class UserThrownError(UserDataFunctionError):
    """Represents an error that is thrown by the user's code during the execution of a function. This error is typically thrown when the user's code encounters an error that is not related to the input data and can be the base class for custom errors."""

    def __init__(self, message: str = "User Exception Thrown", properties: dict = None):
        """
        :param message: An optional message describing the exception. Defaults to "User Exception Thrown".
        :type message: str
        :param properties: An optional dictionary containing additional properties related to the exception.
        :type properties: dict
        """
        super().__init__(UDFExceptionCodes.USER_THROWN, message, properties)



# flake8: noqa: I005
from typing_extensions import deprecated
from fabric.functions.user_data_functions import UserDataFunctions


@deprecated("This class is deprecated. Please use 'UserDataFunctions' instead.")
class FabricApp(UserDataFunctions):
    """
    This class is deprecated. Please use 'UserDataFunctions' instead.
    """

    def __init__(self):
        """
        """
        
        super().__init__()


# flake8: noqa: I005
from typing import Any
from azure.functions.decorators.core import InputBinding
import azure
import json

from fabric.functions.udf_exception import UserDataFunctionInternalError
from fabric.functions.fabric_class import UserDataFunctionContext

class UserDataFunctionContextInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'UserDataFunctionContext'
    
    def __init__(self,
                name: str,
                **kwargs):
        super().__init__(name)


# The input converter that automatically gets registered in the function app.
class UserDataFunctionContextConverter(azure.functions.meta.InConverter, binding='UserDataFunctionContext'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return pytype == UserDataFunctionContext

    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        
        if data is not None and data.type == 'string' and data.value is not None: 
            body = json.loads(data.value) 
        else:
            raise UserDataFunctionInternalError(
                f'Unable to load data successfully for UserDataFunctionContext parameter.')
         
        body = json.loads(data.value) 

        return cls.parseType(body)
    
    @classmethod
    def parseType(self, body: json):

        return UserDataFunctionContext(
                invocationId=body['InvocationId'],
                executingUser=body['ExecutingUser'])    



# flake8: noqa: I005

import datetime
from typing import Any, Optional
import builtins
import json
from fabric.internal.lib_helpers import module_factory
from fabric.functions.udf_exception import UserDataFunctionInvalidInputError

class BasicDatatypeConverter:

    @staticmethod
    def try_convert_type(value: object, pytype: type) -> object:
        if isinstance(value, str):
            # this is the case if the tuple is passed as a string e.g., '[1,2,3]'
            value_as_json = json.loads(value)
            return pytype(value_as_json)
        return pytype(value)
    
    @staticmethod
    def try_convert_bool(value: object) -> bool:
        if value == None:
            return False
        elif isinstance(value, str):
            if value.lower() == 'false' or value == '' or value == '0' or value == '0.0': # Everything but False / false should return True
                return False
            elif value.lower() == 'true' or value == '1' or value == '1.0':
                return True
        elif value == 0 or value == False:
            return False
        elif value == 1 or value == True:
            return True

        raise ValueError("Input is not a valid boolean value")

    @staticmethod
    def try_convert_dataframe(value: object):
        pd = module_factory().load_module('pandas')
        return pd.DataFrame.from_dict(value, orient='columns')

    @staticmethod
    def try_convert_series(value: object):
        pd = module_factory().load_module('pandas')
        return pd.Series(value)

    special_types = {
        "dict": lambda x : BasicDatatypeConverter.try_convert_type(x, dict),
        "list": lambda x : BasicDatatypeConverter.try_convert_type(x, list),
        "set": lambda x : BasicDatatypeConverter.try_convert_type(x, set),
        "tuple": lambda x : BasicDatatypeConverter.try_convert_type(x, tuple),
        "bool": lambda x : BasicDatatypeConverter.try_convert_bool(x),
    }

    @staticmethod
    def check_input_type_annotation(pytype: type) -> bool:
        return True
      
    @staticmethod
    def tryconvert(property_type: str, property_value: str):
        try:
            if property_type == 'datetime':
                # datetime is not a builtin type, so we need to handle it separately
                return datetime.datetime.fromisoformat(property_value)
            elif property_type == "DataFrame":
                # if the value is a string, we need to convert it to a list first
                if isinstance(property_value, str):
                    property_value = json.loads(property_value)
                return BasicDatatypeConverter.try_convert_dataframe(property_value)
            elif property_type == 'Series':
                if isinstance(property_value, str):
                    # if the value is a string, we need to convert it to a list first
                    
                    property_value = json.loads(property_value)
                return BasicDatatypeConverter.try_convert_series(property_value)

            prop_type = get_type(property_type)
            prop_type_name = getattr(prop_type, '__name__', None)
            if prop_type_name in BasicDatatypeConverter.special_types:
                constructor = BasicDatatypeConverter.special_types[prop_type_name]
                return constructor(property_value)
            else:
                return prop_type(property_value)
            
        except (ValueError, TypeError, json.JSONDecodeError): 
            # We can get ValueError, JSONDecodeError, and TypeError due to the special_types that have json.loads
            # Since we can't catch this exception and return it in the formatted output, we will return it as the value of the parameter and then check the parameter types for it
            # We add the parameter name later on in fabric_app.py
            return UserDataFunctionInvalidInputError(properties={'parameter_type': property_type, 'parameter_value': property_value})
        
def get_type(type_name): 
      try: 
          return getattr(builtins, type_name) 
      except AttributeError: 
          return None 


"""You can ignore this module and any files within it. It is used to help set up your User Data Functions.
"""

# flake8: noqa: F401
from .configure_fabric_function_builder import configure_fabric_function_builder


# flake8: noqa: I003
from functools import wraps
from azure.functions import HttpRequest
import inspect
from typing import Callable, TypeVar

from fabric.functions.udf_exception import UserDataFunctionInvalidInputError, UserDataFunctionMissingInputError
from fabric.internal.arrow import arrow_request
from fabric.internal.converters.basic_datatype_converter import BasicDatatypeConverter
from .function_parameter_keywords import REQ_PARAMETER
import json
import asyncio

T = TypeVar('T')

def create_missing_input_error(param: inspect.Parameter):
    annotation_name = getattr(param.annotation, '__name__', None)
    properties = {'parameter_value': "Parameter not found in request"}
    if annotation_name:
        properties['parameter_type'] = annotation_name
    return UserDataFunctionMissingInputError(properties=properties)

def add_parameters(func: Callable[..., T], udfParams: list[inspect.Parameter]):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        # get request from kwargs
        req : HttpRequest = None
        if REQ_PARAMETER in kwargs:
            req = kwargs[REQ_PARAMETER]
        # find parameters in request and add to kwargs
        content_type = req.headers.get('Content-Type')
        if content_type == None or 'application/json' in content_type:
            body = req.get_json()
            for param in udfParams:
                if param.name in body:
                    #kwargs[param.name] = body[param.name]
                    val = body[param.name]
                    kwargs[param.name] = BasicDatatypeConverter.tryconvert(param.annotation.__name__, val)
        elif 'multipart/form-data' in content_type:
            
            # file can be parameters or a octet stream for arrow upload
            for file in req.files:
                
                if file == "parameters":
                    # parse udf parameters
                    json_part = req.files[file]
                    json_data = json.loads(json_part.stream._file.read())
                    for param in udfParams:
                        if param.name in json_data:
                            val = json_data[param.name]
                            kwargs[param.name] = BasicDatatypeConverter.tryconvert(param.annotation.__name__, val)
                        else:
                            kwargs[param.name] = create_missing_input_error(param)
                else:
                    # ensure parameter exists
                    param = next((item for item in udfParams if item.name == file), None)
                    if param is None:
                        continue
                    # parse as arrow to pandas
                    # file = name of part = name of parameter
                    try:
                        arrow_part = req.files[file].read()
                        ar_request = arrow_request(arrow_part)
                        pandas_df = ar_request.to_pandas()
                        kwargs[file] = pandas_df
                    except Exception as e:
                        kwargs[arg.name] = UserDataFunctionInvalidInputError(properties={'parameter_type': param.annotation.__name__, 'parameter_value': "Unable to parse"})
                continue
            
        for arg in udfParams:
            if arg.name not in kwargs:
                kwargs[arg.name] = create_missing_input_error(arg)

        if asyncio.iscoroutinefunction(func):
            return await func(*args, **kwargs) 
        else: 
            return func(*args, **kwargs) 
    return wrapper


import json
from typing import Any, Optional

# flake8: noqa: I900
import azure
from azure.functions.decorators.core import DataType, InputBinding

from fabric.functions.fabric_item import FabricItem
from fabric.internal.fabric_lakehouse_files_client import FabricLakehouseFilesClient
from fabric.functions.udf_exception import UserDataFunctionInternalError

from fabric.functions.fabric_class import (FabricSqlConnection, FabricLakehouseClient, FabricVariablesClient)

# The binding object that will be used by our input decorator below
class FabricItemInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'FabricItem'
    
    def __init__(self,
                name: str,
                alias: Optional[str] = None,
                audienceType: Optional[str] = None,
                **kwargs):
        super().__init__(name, DataType.STRING)

# The input converter that automatically gets registered in the function app. 
class FabricItemConverter(azure.functions.meta.InConverter,
                              binding='FabricItem'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:

        # Need to check what the enum type of the item is and make sure we can convert it
        return issubclass(pytype, (FabricItem, FabricSqlConnection))
    
    @classmethod
    def parseType(self, body: json):
        endpoints = body['Endpoints']
        endpoints = {k.lower(): v for k, v in endpoints.items()}  # we can know this doesn't have collisions 
        # because the host extension uses a case insensitive dictionary
        sqlEndpoint = "sqlendpoint"
        fileEndpoint = "fileendpoint"
        # endpoint keys are normalized to lowercase above, use lowercase key here
        vlEndpoint = "vlendpoint"
        

        if sqlEndpoint in endpoints and fileEndpoint in endpoints:
            return FabricLakehouseClient(
                alias_name=body['AliasName'],
                endpoints=endpoints,
            )
        elif sqlEndpoint in endpoints:
            return FabricSqlConnection(
                alias_name=body['AliasName'],
                endpoints=endpoints,
            )
        elif fileEndpoint in endpoints:
            return FabricLakehouseFilesClient(
                alias_name=body['AliasName'],
                endpoints=endpoints,
            )
        elif vlEndpoint in endpoints:
            # Validate expected keys to provide clearer errors during local testing
            vl_info = endpoints.get(vlEndpoint, {})
            if not isinstance(vl_info, dict) or 'ConnectionString' not in vl_info or 'AccessToken' not in vl_info:
                raise UserDataFunctionInternalError(f"Invalid vlendpoint payload for alias {body.get('AliasName')}")
            return FabricVariablesClient(
                alias_name=body['AliasName'],
                endpoints=endpoints,
            )

        # If not found above, return the default
        return FabricItem(
            alias_name=body['AliasName'],
            endpoints=endpoints)
    
    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        if data is not None and data.type == 'string' and data.value is not None: 
            body = json.loads(data.value)
            error = body.get('ErrorMessage', None)
            if error:
                return UserDataFunctionInternalError(
                    f'Unable to load data successfully for fabric item', {'reason': error})
        else:
            return UserDataFunctionInternalError(
                f'Unable to load data successfully for fabric item')

        return cls.parseType(body)


# flake8: noqa: I003
import inspect
from typing import Any, Callable, List, Optional, Type
from azure.functions import HttpRequest, HttpResponse, FunctionApp, Context
from fabric.functions.fabric_class import (
    FabricLakehouseClient,
    FabricSqlConnection,
    UserDataFunctionContext,
    FabricVariablesClient,
)
from fabric.functions.fabric_item import FabricItem
from fabric.internal.decorators.constants import SpecConstants
from fabric.internal.utils.spec_utils import get_multipart_content
from fabric.internal.openapi_spec_generator import (
    create_pydantic_model,
    FunctionMetadata
)
from fabric.internal.utils.docstring_parser import extract_summary_description
from .ensure_formatted_returntype import ensure_formatted_returntype
from .add_timeout import add_timeout
from .remove_unused_binding_params import remove_unused_binding_params
from .add_parameters import add_parameters
from .log_error import log_error
from fabric.internal.fabric_lakehouse_files_client import FabricLakehouseFilesClient
from pydantic import BaseModel

from .function_parameter_keywords import (
    REQ_PARAMETER,
    CONTEXT_PARAMETER,
    UNUSED_FABRIC_CONTEXT_PARAMETER,
)


def configure_fabric_function_builder(
    udf: FunctionApp, wrap, functions_metadata: List[FunctionMetadata]
) -> Callable[..., Any]:
    def decorator(func):
        sig = inspect.signature(func)

        # Update function parameters to include a request object for validation
        params = []
        params.append(inspect.Parameter(REQ_PARAMETER, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=HttpRequest))
        params.append(inspect.Parameter(UNUSED_FABRIC_CONTEXT_PARAMETER, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=UserDataFunctionContext))
        params.append(inspect.Parameter(CONTEXT_PARAMETER, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Context))

        # Udf params to be parsed from the request
        udfParams = []
        for param in sig.parameters.values():
            # Ensure bindings are still there
            if _is_typeof_fabricitem_input(
                param.annotation
            ) or _is_typeof_userdatafunctioncontext_input(param.annotation):
                params.append(
                    inspect.Parameter(
                        param.name,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        annotation=param.annotation,
                    )
                )
            # Separate out basic parameters to parse later
            if param.name != REQ_PARAMETER and param.name != CONTEXT_PARAMETER:
                udfParams.append(
                    inspect.Parameter(
                        param.name,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        annotation=_get_cleaned_type_and_wrap_str(param),
                    )
                )
        sig = sig.replace(parameters=tuple(params)).replace(return_annotation=str)
        func.__signature__ = sig
        annotations = {}
        # Update annotations to ensure it uses the cleaned type
        for param in params:
            annotations[param.name] = param.annotation
        # Update return annotation of func to be HttpResponse
        # We should catch if they don't have a return type during metadata generation, but good to double check here
        if "return" in func.__annotations__:
            annotations["old_return"] = func.__annotations__["return"]

        annotations["return"] = HttpResponse
        func.__annotations__ = annotations

        # Add wrapper for function to handle ensure all return values are parsed to HttpResponse
        user_func = log_error(func)
        user_func = remove_unused_binding_params(user_func)
        user_func = add_timeout(user_func)
        user_func = ensure_formatted_returntype(user_func)

        # Add parameters to the function
        user_func = add_parameters(user_func, udfParams)
        fb = udf._validate_type(user_func)
        udf._function_builders.append(fb)

        _build_spec(user_func, udfParams, functions_metadata, func.__annotations__)

        return wrap(fb, user_func)

    return decorator


def _is_typeof_fabricitem_input(obj):
    # Check to see if parameter is anything we might return from a fabric binding
    return (
        obj == FabricSqlConnection
        or obj == FabricLakehouseFilesClient
        or obj == FabricLakehouseClient
        or obj == FabricItem
        or obj == FabricVariablesClient
    )


def _is_typeof_userdatafunctioncontext_input(obj):
    # Check to see if parameter is anything we might return from a fabric binding
    return obj == UserDataFunctionContext


def _get_cleaned_type_and_wrap_str(param):
    if hasattr(param.annotation, "__origin__"):
        return param.annotation.__origin__
    else:
        return param.annotation
    
def _build_spec(
    user_func: Callable[..., Any],
    udfParams: List[inspect.Parameter],
    functions_metadata: List[FunctionMetadata],
    func_annotations: dict,
) -> None:
    # Remove FabricSqlConnection, FabricLakehouseClient, UserDataFunctionContext from the parameters
    input_params: List[inspect.Parameter] = [
        param for param in udfParams if param.annotation not in [
            FabricSqlConnection, 
            FabricLakehouseClient, 
            UserDataFunctionContext,
            FabricItem,
            FabricVariablesClient
        ]
    ]


    # Create a pydantic model from udfParams
    req_param_model_name: str = f"{user_func.__name__.capitalize()}RequestModel"
    req_param_model: Type[BaseModel] = create_pydantic_model(f"{req_param_model_name}", input_params)
    # Create a pydantic model from the return type
    res_param_model_name: str = f"{user_func.__name__.capitalize()}ResponseModel"
    old_return_annotation: Optional[Type] = func_annotations.get("old_return", None)
    
    # Create full response model with all fields
    res_param_model: Type[BaseModel] = create_pydantic_model(
        f"{res_param_model_name}",
        [
            inspect.Parameter(
                "functionName",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            ),
            inspect.Parameter(
                "invocationId",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            ),
            inspect.Parameter(
                "status",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            ),
            inspect.Parameter(
                "output",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=old_return_annotation,
            ),
            inspect.Parameter(
                "errors",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=List[str],
                default=[],
            )
        ],
    )

    if user_func.__name__ not in [SpecConstants.SPEC_FUNCTION_NAME_JSON, SpecConstants.SPEC_FUNCTION_NAME_YAML]:
        summary: str
        desc: str
        summary, desc = extract_summary_description(user_func.__doc__)
        multipart_req: dict
        binary_res: dict
        multipart_req_model: Optional[Type[BaseModel]] = None
        multipart_req, binary_res, multipart_req_model = get_multipart_content(user_func.__name__, udfParams)
        func_metadata: FunctionMetadata = FunctionMetadata(user_func.__name__)
        func_metadata.add_path(
            f"/{user_func.__name__}/invoke",
            {
                "post": {
                    "summary": f"{summary}",
                    "description": f"{desc}",
                    "operationId": f"{user_func.__name__}",
                    "requestBody": {
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": f"#/components/schemas/{req_param_model.__name__}"
                                }
                            },
                            **multipart_req
                        }
                    },
                    "responses": {
                        "200": {
                            "description": "Successful response",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "$ref": f"#/components/schemas/{res_param_model.__name__}"
                                    }
                                },
                                **binary_res
                            },
                        },
                        "400": {
                            "description": "Bad Request - Invalid input, missing input, or user thrown error",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "401": {
                            "description": "Unauthorized",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "403": {
                            "description": "Forbidden",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "408": {
                            "description": "Request Timeout - Execution time exceeded timeout limit",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "409": {
                            "description": "Conflict",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "413": {
                            "description": "Response Too Large - Response exceeds the maximum size limit",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        },
                        "500": {
                            "description": "Internal Server Error - System exception or programming error",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                    }
                                }
                            }
                        }
                    },
                }
            },
        )
        func_metadata.add_component(req_param_model_name, req_param_model)
        func_metadata.add_component(res_param_model_name, res_param_model)
        
        if multipart_req_model:
            func_metadata.add_component(multipart_req_model.__name__, multipart_req_model)

        functions_metadata.append(func_metadata)


class Timeout:

    FUNC_TIMEOUT_IN_SECONDS = 200


class SizeLimit:

    RESPONSE_SIZE_LIMIT_IN_MB = 30


class SpecConstants:
    LOCAL_HOST_URL = "http://localhost:7071/api/"
    PUBLIC_URL_DESC = "The public URL to invoke the User Data Function."
    SPEC_FUNCTION_NAME_JSON = "get_udf_oai_spec_json"
    SPEC_FUNCTION_NAME_YAML = "get_udf_oai_spec_yaml"


class UDFExceptionCodes:
    INVALID_INPUT = "InvalidInput"
    MISSING_INPUT = "MissingInput"
    RESPONSE_TOO_LARGE = "ResponseTooLarge"
    TIMEOUT = "Timeout"
    USER_THROWN = "UserThrown"
    INTERNAL_ERROR = "InternalError"



import json

# flake8: noqa: R505
class StatusCode:
    BAD_REQUEST = "BadRequest"
    TIMEOUT = "Timeout"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    RESPONSE_TOO_LARGE = "ResponseTooLarge"


class FormattedError:
    def __init__(self, error_code: str, message: str, properties: dict = None):
        self.errorCode: str = error_code
        self.message: str = message
        if properties is None:
            self.properties = {}
        else:
            self.properties = properties

    def add_or_update_property(self, key, value):
        self.properties[key] = value

    def to_json(self):
        return self.__dict__


class UserDataFunctionInvokeResponse:
    def __init__(self):
        self.functionName: str = ""
        self.invocationId: str = ""
        self.status: StatusCode = StatusCode.SUCCEEDED
        self.output: object = ""
        self.errors: list[FormattedError] = []

    def add_error(self, error):
        self.errors.append(error)

    def to_json(self):
        def convert(o):
            if isinstance(o, StatusCode):
                return o.value
            elif isinstance(o, FormattedError):
                return o.to_json()
            return o.__dict__

        return json.dumps(
            self,
            default=convert)



# Since str can't have a binding different than the Generic binding, we change its annotation to this class so it can be bound differently
class UdfString:
    def __init__(self, value: str):
        self.value = value

    def __name__(self):
        return 'UdfString'

    def __str__(self):
        return self.value



# flake8: noqa: I003
from functools import wraps
from typing import Callable, TypeVar
from fabric.internal.decorators.function_parameter_keywords import UNUSED_FABRIC_CONTEXT_PARAMETER, REQ_PARAMETER
import asyncio

T = TypeVar('T')

def remove_unused_binding_params(func: Callable[..., T]):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        if REQ_PARAMETER in kwargs:
            del kwargs[REQ_PARAMETER]
        if UNUSED_FABRIC_CONTEXT_PARAMETER in kwargs:
            del kwargs[UNUSED_FABRIC_CONTEXT_PARAMETER]

        if asyncio.iscoroutinefunction(func):
            return await func(*args, **kwargs) 
        else: 
            return func(*args, **kwargs) 
    return wrapper


# flake8: noqa: I003

from datetime import datetime, timezone, timedelta
import logging
from typing import Generator
from azure.cosmos import CosmosClient
from azure.core.credentials import TokenCredential
from fabric.functions.fabric_item import FabricItem

from threading import Lock

_client_iters = {}
_client_iters_lock = Lock()

def get_cosmos_client(item: FabricItem, cosmos_db_uri: str) -> CosmosClient:
    with _client_iters_lock:
        if cosmos_db_uri not in _client_iters:
            _client_iters[cosmos_db_uri] = _create_cosmosclient_getter(cosmos_db_uri)

        client_iter = _client_iters[cosmos_db_uri]

    return client_iter.send(item)

def _cosmosclient_getter(cosmos_db_uri: str) -> Generator[CosmosClient, FabricItem, None]:
        cosmos_client: CosmosClient = None
        client_cred: TokenCredential = None
        refresh_after: datetime = None
        try:
            # pump
            item: FabricItem = yield

            while True:
                if refresh_after is not None and refresh_after >= datetime.now(timezone.utc) and cosmos_client is not None:
                    logging.info('Reusing existing CosmosClient')
                    pass
                else:
                    credential = item.get_access_token()
                    client_cred = credential
                    refresh_after = datetime.fromtimestamp(credential.get_token().expires_on, tz=timezone.utc) - timedelta(minutes=1)
                    logging.info('Creating new CosmosClient')
                    cosmos_client = CosmosClient(cosmos_db_uri, client_cred)
                item = yield cosmos_client
        except GeneratorExit as e:
            logging.debug(f'CosmosClient generator closed: {e}')
            pass

def _create_cosmosclient_getter(cosmos_db_uri: str) -> Generator[CosmosClient, FabricItem, None]:
    getter = _cosmosclient_getter(cosmos_db_uri)
    next(getter)
    return getter


# flake8: noqa : I005, I900, I001, I003

from fabric.internal.openapi_spec_generator import create_pydantic_model
from pydantic import BaseModel
import copy
import inspect


def get_multipart_content(
    func_name: str, udf_params: list[inspect.Parameter]
) -> tuple[dict, dict, BaseModel]:
    # If udfParams contains a parameter of type pandas.DataFrame, then the request is expected to be a multipart/form-data request
    # where the parameters are in the 'parameters' part and the files are in the 'files' part
    udf_params_copy = copy.deepcopy(udf_params)
    has_pandas_df = False
    df_params = list(
        filter(
            lambda param: param.annotation.__name__ == "DataFrame"
            or param.annotation.__name__ == "Series",
            udf_params_copy,
        )
    )
    arrow_params = {}
    if len(df_params) > 0:
        has_pandas_df = True
        for param in df_params:
            # Remove the parameter from the list of parameters
            udf_params_copy.remove(param)
            # Add the parameter to the arrow_params dictionary
            arrow_params[param.name] = {
                "type": "string",
                "format": "binary",
                "description": f"The Arrow file for the parameter '{param.name}'",
            }

    if len(arrow_params) > 0:
        req_param_model_name = f"{func_name.capitalize()}MultipartRequestModel"
        req_param_model = create_pydantic_model(
            f"{req_param_model_name}", udf_params_copy
        )
    else:
        req_param_model = {}
    # If the function has a pandas.DataFrame parameter, then the request is expected to be a multipart/form-data request
    # where the parameters are in the 'parameters' part and the files are in the 'files' part
    if has_pandas_df:
        req = {
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "required": ["parameters", *arrow_params.keys()],
                    "properties": {
                        "parameters": {
                            "$ref": f"#/components/schemas/{req_param_model.__name__}"
                        },
                        **arrow_params,
                    },
                },
                "encoding": {
                    "parameters": {"contentType": "application/json"},
                    **{
                        param: {"contentType": "application/octet-stream"}
                        for param in arrow_params.keys()
                    },
                },
            }
        }
        res = {
            "application/octet-stream": {
                "schema": {"type": "string", "format": "binary"}
            }
        }
        return req, res, req_param_model
    else:
        return {}, {}, None




import logging


class UdfLogger(logging.Logger):

    def __init__(self, name: str):
        super().__init__(name)

    def error(self, msg, *args, **kwargs):
        msg = f"Python Worker: {msg}"
        super().error(msg, *args, **kwargs)



"""This is the module that contains all of the relevant files you will need to import.
For example, 'import fabric.functions' or 'import fabric.functions as fn'.
"""

# flake8: noqa: F401
from .fabric_app import FabricApp
from .fabric_class import FabricSqlConnection, FabricLakehouseClient, UserDataFunctionContext, FabricVariablesClient
from .udf_exception import UserDataFunctionError, UserDataFunctionInternalError, UserDataFunctionInvalidInputError, UserDataFunctionMissingInputError, UserDataFunctionResponseTooLargeError, UserDataFunctionTimeoutError, UserThrownError
from .user_data_functions import UserDataFunctions
from .fabric_item import FabricItem
from .generic_audience_types import GenericAudienceTypes

__all__ = [
    "UserDataFunctions",
    "FabricSqlConnection",
    "FabricLakehouseClient",
    "FabricVariablesClient",
    "UserDataFunctionContext",
    "UserDataFunctionError",
    "UserDataFunctionInternalError",
    "UserDataFunctionInvalidInputError",
    "UserDataFunctionMissingInputError",
    "UserDataFunctionResponseTooLargeError",
    "UserDataFunctionTimeoutError",
    "UserThrownError",
    "FabricItem",
    "GenericAudienceTypes",
]



# flake8: noqa: I005
from logging import Logger
import typing

from azure.functions import AppExtensionBase, Context, HttpResponse

INVOCATION_ID_HEADER = 'x-ms-invocation-id'
INVOCATION_ID_PARAMETER = 'reqInvocationId'

class InvocationIdMiddleware(AppExtensionBase):
    """A Python worker extension to add the invocation id to the response headers.
    """

    @classmethod
    def init(cls):
        pass

    @classmethod
    def configure(cls, *args, append_to_http_response:bool=False, **kwargs):
        pass

    @classmethod
    def pre_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: typing.Dict[str, object],
        *args, **kwargs
    ) -> None:
        func_args.update({INVOCATION_ID_PARAMETER : context.invocation_id})

    @classmethod
    def post_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: typing.Dict[str, object],
        func_ret: typing.Optional[object],
        *args, **kwargs
    ) -> None:
        if func_ret.headers is not None:
            func_ret.headers.add(INVOCATION_ID_HEADER, context.invocation_id)


# flake8: noqa: I005
import pyarrow as pa

class arrow_series_response():

    def __init__(self, series):
        self._series = series

    def to_bytes(self) -> bytes:
        sink = pa.BufferOutputStream()
        arrow_array = pa.array(self._series)
        arrow_name = self._series.name if self._series.name else "data"
        writer = pa.ipc.new_stream(sink, schema=pa.schema([(arrow_name, arrow_array.type)]))
        batch_arrow = pa.record_batch([arrow_array], names=[arrow_name])
        writer.write_batch(batch_arrow)
        writer.close()
        return sink.getvalue().to_pybytes()



# flake8: noqa: I003

from functools import wraps
from azure.functions import HttpResponse
import json
import inspect
import sys
from typing import Any, Callable, TypeVar
import fabric.functions.udf_exception as udf_exceptions
from fabric.internal.arrow import arrow_dataframe_response, arrow_series_response
from fabric.internal.invoke_response import FormattedError, StatusCode, UserDataFunctionInvokeResponse
from fabric.internal.middleware._invocationIdMiddleware import INVOCATION_ID_PARAMETER
from .function_parameter_keywords import REQ_PARAMETER

from fabric.internal.logging import UdfLogger

from .constants import SizeLimit

import asyncio

logger = UdfLogger(__name__)

T = TypeVar('T')

def ensure_formatted_returntype(func: Callable[..., T], response_size_limit_in_mb: int = SizeLimit.RESPONSE_SIZE_LIMIT_IN_MB):
    def _check_args_for_exceptions(args: tuple, kwargs: dict):
        exceptions = []
        for arg in args:
            if issubclass(type(arg), udf_exceptions.UserDataFunctionError):
                exceptions.append(arg)
        for key, value in kwargs.items():
            if issubclass(type(value), udf_exceptions.UserDataFunctionError):
                value.properties['parameter_name'] = key
                exceptions.append(value)
        
        return exceptions
    
    def _ensure_response_is_not_too_large(resp: Any):
        if (sys.getsizeof(resp) > response_size_limit_in_mb * 1024 * 1024): # X MB = 1024 (bytes) * 1024 (kilobytes) * X
            raise udf_exceptions.UserDataFunctionResponseTooLargeError(response_size_limit_in_mb)

    def _is_pandas_type(obj) -> bool:
        # Check if the object is a pandas dataframe or series
        obj_type = inspect.getmodule(obj)
        if obj_type is not None:
            return 'pandas.core' in obj_type.__package__
        return False

    def _ensure_response_is_json_serializable(resp: Any):
        ret = ""
        try:
            if _is_pandas_type(resp):
                # If it's a dataframe/series we need to convert it to a dictionary
                if resp.__class__.__name__ == 'DataFrame':
                    resp = resp.to_dict(orient='records')
                else:
                    resp = resp.to_dict() # Adding Orient=type will break pandas.Series as this is not a valid argument for to_dict
                _ensure_response_is_not_too_large(resp)  
            else:
                output = json.dumps(resp) # we are checking it is json serializable here, not that we actually need the value of it
                # call sys.getsizeof on the string representation, since we know sys.getsizeof will work on the string representation
                _ensure_response_is_not_too_large(output)     
            ret = resp

        except (TypeError, OverflowError):
            ret = getattr(resp, '__dict__', str(resp))
            _ensure_response_is_not_too_large(ret)
            
        return ret
    
    def _ensure_response_is_binary_serializeable(resp: Any):
        ret = ""
        try:
            arrow_type = None
            if resp.__class__.__name__ == 'DataFrame':
                arrow_type = arrow_dataframe_response(resp)
            else:
                arrow_type = arrow_series_response(resp)

            arrow_bytes = arrow_type.to_bytes()
            # call sys.getsizeof on the string representation, since we know sys.getsizeof will work on the string representation
            _ensure_response_is_not_too_large(arrow_bytes)     
            ret = arrow_bytes

        except (TypeError, OverflowError):
            ret = getattr(resp, '__dict__', str(resp))
            _ensure_response_is_not_too_large(ret)
            
        return ret

    def _log_and_convert_to_formatted_error(e: Exception):
            ret = FormattedError(getattr(e, "error_code", type(e).__name__), getattr(e, 'message', str(e)), getattr(e, 'properties', {}))

            logger.error(f"Error during function invoke: {ret.to_json()}")
            return ret
        
    @wraps(func)
    async def wrapper(*args, **kwargs):
        invocationId = kwargs[INVOCATION_ID_PARAMETER]
        del kwargs[INVOCATION_ID_PARAMETER]

        # Use request to know how to format the return type
        req = kwargs[REQ_PARAMETER]

        invoke_response = UserDataFunctionInvokeResponse()
        invoke_response.functionName = func.__name__
        invoke_response.invocationId = invocationId

        try:
            input_exceptions = _check_args_for_exceptions(args, kwargs)
            if len(input_exceptions) > 0:
                invoke_response.status = StatusCode.BAD_REQUEST
                for exception in input_exceptions:
                    invoke_response.add_error(_log_and_convert_to_formatted_error(exception))
            else:
                # The line that actually invokes the user's function
                if asyncio.iscoroutinefunction(func):
                    resp = await func(*args, **kwargs) 
                else: 
                    resp = func(*args, **kwargs) 

                # Find return value in annotations. If is a dataframe or series and request is formdata
                if 'multipart/form-data' in req.headers.get('Content-Type') and _is_pandas_type(resp):
                    # If it's a data frame and it's successful we need to convert it to bytes and return right away
                    serializable_resp = _ensure_response_is_binary_serializeable(resp)
                    invoke_response.content_type = "application/octet-stream"
                    headers = {
                        'x-fabric-udf-status': str(StatusCode.SUCCEEDED)
                    }
                    formatted_response = {"status_code": 200, "headers": headers, "mimetype": "application/octet-stream", "charset": "utf-8"}
                    return HttpResponse(body=serializable_resp, status_code=formatted_response['status_code'], headers=formatted_response['headers'], mimetype=formatted_response['mimetype'], charset=formatted_response['charset'])

                # If it's not a dataframe we need to run existing checks
                serializable_resp = _ensure_response_is_json_serializable(resp)

                invoke_response.output = serializable_resp
                invoke_response.status = StatusCode.SUCCEEDED
        except Exception as e:
            if issubclass(type(e), udf_exceptions.UserDataFunctionError):
                invoke_response.add_error(_log_and_convert_to_formatted_error(e))

                if type(e) is udf_exceptions.UserDataFunctionTimeoutError:
                    invoke_response.status = StatusCode.TIMEOUT
                elif type(e) is udf_exceptions.UserDataFunctionResponseTooLargeError:
                    invoke_response.status = StatusCode.RESPONSE_TOO_LARGE
                elif issubclass(type(e), udf_exceptions.UserThrownError): # custom exceptions that the user can throw, or they can use as a base class
                    invoke_response.status = StatusCode.BAD_REQUEST
                else: # custom exceptions that we  throw
                    invoke_response.status = StatusCode.FAILED
            else:
                invoke_response.status = StatusCode.FAILED
                # Put the details into an InternalErrorException to hide other details from the exception
                error = udf_exceptions.UserDataFunctionInternalError(properties={'error_type': type(e).__name__, 'error_message': getattr(e, 'message', str(e))})
                invoke_response.add_error(_log_and_convert_to_formatted_error(error))

        # we need to make a new HttpResponse because there is no way to modify the body of an existing one
        headers = {
            'x-fabric-udf-status': str(invoke_response.status)
        }
        formatted_response = {"status_code": 200, "headers": headers, "mimetype": "application/json", "charset": "utf-8"}
        return HttpResponse(body=invoke_response.to_json(), status_code=formatted_response['status_code'], headers=formatted_response['headers'], mimetype=formatted_response['mimetype'], charset=formatted_response['charset'])

    return wrapper



# flake8: noqa: I003

import asyncio
from functools import wraps
import inspect
import sys
import logging
import traceback
from typing import Callable, TypeVar
T = TypeVar('T')

def log_error(func: Callable[..., T]):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            if inspect.iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        except Exception as e:
            stacktrace = traceback.format_exc()
            logging.error(stacktrace)
            raise
        
    return wrapper


import logging

# flake8: noqa: I900
from fabric.metadata.metadata_generator import (
    ERROR_CODE, SUCCESS, _collect_function_metadata, _get_app_functions)


def validate():
    try:
        functions = _get_app_functions()
        if(functions is None):
            return ERROR_CODE
        
        metadata = _collect_function_metadata(functions)
        if(metadata is None):
            return ERROR_CODE
        
        return SUCCESS
    except Exception as e:
        logging.error(f'Got errors when generating metadata. Details:{e}')
        return ERROR_CODE


if __name__ == '__main__':
    validate()



# flake8: noqa: I005
import jwt
from azure.core.credentials import AccessToken, TokenCredential  # noqa: I900


class CustomTokenCredential(TokenCredential):
    def __init__(self, token: str):
        alg = jwt.get_unverified_header(token)['alg']
        decoded_access_token = jwt.decode(token, algorithms=[alg], options={"verify_signature": False,  # Needed or else it complains about Exception: ValueError: Unable to load PEM file.
                                                                            "verify_exp": False}  # Needed to not throw if the access token is already expired
                                          )
        # Token Expiry
        token_expiry = decoded_access_token["exp"]

        self._token = AccessToken(token, token_expiry)

    def get_token(self, *scopes, **kwargs) -> AccessToken:
        return self._token



# flake8: noqa: I003
from functools import wraps
import queue
import threading
import time
import traceback
from typing import Callable, Optional, TypeVar

from fabric.functions.udf_exception import UserDataFunctionTimeoutError
from fabric.internal.logging import UdfLogger
from fabric.internal.decorators.function_parameter_keywords import CONTEXT_PARAMETER

from .constants import Timeout

import asyncio
import inspect
import functools
import contextvars

T = TypeVar('T')

logger = UdfLogger(__name__)

def add_timeout(func: Callable[..., T], function_timeout: int = Timeout.FUNC_TIMEOUT_IN_SECONDS):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        context = None
        # Extract Azure Function Context to setup invocation id for logging
        if CONTEXT_PARAMETER in kwargs:
            context = kwargs[CONTEXT_PARAMETER]
            del kwargs[CONTEXT_PARAMETER]

        try:
            # This will raise a TimeoutError if the function takes longer than the timeout
            loop = asyncio.get_running_loop()
            func_task = None

            if inspect.iscoroutinefunction(func):
                # For async functions, create a task in the current event loop
                # Set up context in the current thread before creating the task
                if context is not None:
                    context.thread_local_storage.invocation_id = context.invocation_id
                
                func_task = asyncio.create_task(func(*args, **kwargs))
            else:
                # For sync functions, run in thread pool with proper context setup
                def sync_function_runner():
                    if context is not None:
                        context.thread_local_storage.invocation_id = context.invocation_id
                    return func(*args, **kwargs)
                ctx = contextvars.copy_context()
                func_task = loop.run_in_executor(None, ctx.run, sync_function_runner)

            return await asyncio.wait_for(func_task, timeout=function_timeout)

        except asyncio.TimeoutError:
            return UserDataFunctionTimeoutError(function_timeout)
        
    return wrapper


# flake8: noqa: I005
import datetime
from typing import Any, Optional
from azure.functions.decorators.core import DataType, InputBinding
import azure
import json
import builtins

from fabric.functions.udf_exception import UserDataFunctionInternalError, UserDataFunctionInvalidInputError, UserDataFunctionMissingInputError

class UdfPropertyInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'UdfProperty'
    
    def __init__(self,
                name: str,
                parameterName: str,
                typeName: Optional[str] = None,
                data_type: Optional[DataType] = DataType.STRING,
                **kwargs):
        super().__init__(name, data_type)


# The input converter that automatically gets registered in the function app.
class UdfPropertyConverter(azure.functions.meta.InConverter, binding='UdfProperty'):
    
    def try_convert_type(value: object, pytype: type) -> object:
        if isinstance(value, str):
            # this is the case if the tuple is passed as a string e.g., '[1,2,3]'
            value_as_json = json.loads(value)
            return pytype(value_as_json)
        return pytype(value)
    
    def try_convert_bool(value: object) -> bool:
        if value == None:
            return False
        elif isinstance(value, str):
            if value.lower() == 'false' or value == '' or value == '0' or value == '0.0': # Everything but False / false should return True
                return False
            elif value.lower() == 'true' or value == '1' or value == '1.0':
                return True
        elif value == 0 or value == False:
            return False
        elif value == 1 or value == True:
            return True

        raise ValueError("Input is not a valid boolean value")

    special_types = {
        "dict": lambda x : UdfPropertyConverter.try_convert_type(x, dict),
        "list": lambda x : UdfPropertyConverter.try_convert_type(x, list),
        "set": lambda x : UdfPropertyConverter.try_convert_type(x, set),
        "tuple": lambda x : UdfPropertyConverter.try_convert_type(x, tuple),
        "bool": lambda x : UdfPropertyConverter.try_convert_bool(x),
    }

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return True



    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        if data is not None and data.type == 'string' and data.value is not None:
            body = json.loads(data.value)
            error = body.get('ErrorMessage', None)
            if error:
                if error == "Parameter does not exist in binding data":
                    return UserDataFunctionMissingInputError()
                
                return UserDataFunctionInternalError(
                    f"Unable to load data successfully for parameter. This could be because the invocation isn't valid JSON.", {'reason': error})
        else:
            return UserDataFunctionInternalError(
                'Unable to load data successfully for udf property')
        return tryconvert(body['PropertyType'], body['PropertyJsonValue'])

def get_type(type_name): 
      try: 
          return getattr(builtins, type_name) 
      except AttributeError: 
          return None 

def tryconvert(property_type: str, property_json_value: str):
    if not property_json_value:  # This can happen if/when the value is empty, so we should return None
        return None
    value_as_json = json.loads(property_json_value)
    if value_as_json == None:
        return None
    
    if property_type == 'datetime': # datetime is not a built in, so we must handle it separately
        return datetime.datetime.fromisoformat(value_as_json)

    if property_type == 'UdfString': # locate can't find custom classes within its
        # own module unless their names are qualified with module name, so let's 
        # just make the only custom type we are allowing in a conditional here
        return value_as_json

    # Built ins go here
    prop_type = get_type(property_type)
    prop_type_name = getattr(prop_type, '__name__', None)
    try:
        if prop_type_name in UdfPropertyConverter.special_types:
            constructor = UdfPropertyConverter.special_types[prop_type_name]
            return constructor(value_as_json)
        return prop_type(value_as_json)
    except (ValueError, TypeError, json.JSONDecodeError): 
        # We can get ValueError, JSONDecodeError, and TypeError due to the special_types that have json.loads
        # Since we can't catch this exception and return it in the formatted output, we will return it as the value of the parameter and then check the parameter types for it
        # We add the parameter name later on in fabric_app.py
        return UserDataFunctionInvalidInputError(properties={'parameter_type': property_type, 'parameter_value': value_as_json})



# flake8: noqa: I005

import pyarrow as pa

class arrow_request():

    def __init__(self, arrow_bytes):
        self._arrow_bytes = arrow_bytes

    def to_pandas(self):
        arrow_buffer = pa.BufferReader(self._arrow_bytes)
        table = pa.ipc.open_stream(arrow_buffer).read_all()  # Use the buffer here
        return table.to_pandas()



import typing
from fabric.internal.custom_token_credential import CustomTokenCredential 
from azure.core.credentials import TokenCredential  # noqa: I900

# flake8: noqa: I900
class FabricItem:
    def __init__(self, alias_name: str, endpoints: typing.Dict[str, typing.Dict[str, str]]):
        """Don't worry about using the constructor to create an instance of this class. Fabric will automatically create it for you (as long as you follow the steps within `Remarks`).

        :param alias_name: The alias for the data source being connected to, configured in the portal.
        :type alias_name: str
        :param endpoints: The different endpoints for the data source.
        :type endpoints: typing.Dict[str, typing.Dict[str, str]]
        """
        self.__alias_name = alias_name
        self.__endpoints = endpoints

    @property
    def alias_name(self) -> typing.Optional[str]:
        return self.__alias_name

    @property
    def endpoints(self) -> typing.Optional[typing.Dict[str, typing.Dict[str, str]]]: # noqa TAE002
        return self.__endpoints
    
    def get_access_token(self) -> TokenCredential:
        """Get the access token for the Fabric item.

        :return: The access token.
        :rtype: TokenCredential
        """
        token = self.__endpoints[next(iter(self.__endpoints))]["AccessToken"]

        return CustomTokenCredential(token)



"""
This module defines the UserDataFunctions class, which is used to define User Data Functions in Fabric.
You can use this class by importing it from the fabric.functions module
(e.g. `from fabric.functions import UserDataFunctions`, or `import fabric.functions as fn` -> `fn.UserDataFunctions()`).
"""

# flake8: noqa: I005
import copy
import os
from typing import Any, Callable, Optional
from azure.functions.decorators.http import HttpTrigger, HttpOutput, HttpMethod
from azure.functions.decorators.core import AuthLevel
from azure.functions import FunctionApp
import fabric.internal.decorators as fabric_decorators
from fabric.internal.decorators.constants import SpecConstants
from typing_extensions import deprecated

from fabric.internal.item_binding import FabricItemInput

from fabric.internal.user_data_function_context_binding import (
    UserDataFunctionContextInput,
)
from fabric.internal.logging import UdfLogger
from fabric.internal.openapi_spec_generator import (
    FunctionMetadata,
    OpenAPISpecification,
    OpenAPISpecificationBuilder,
)

from typing import Callable, TypeVar
import importlib

T = TypeVar("T")

# This will prevent a user from writing a function with a 'context' parameter
CONTEXT_PARAMETER = "context"
UNUSED_FABRIC_CONTEXT_PARAMETER = "notusedfabriccontext"
REQ_PARAMETER = "req"


class UserDataFunctions(FunctionApp):
    """
    This class is necessary to define User Data Functions in Fabric. Please ensure an instantiation of this class exists in your code before any User Data Functions are defined.

    .. remarks::
        This class is used to define User Data Functions in Fabric. The class must be instantiated before any User Data Functions are defined. The instantiation of this class is required.

        .. code-block:: python
            import fabric.functions as fn

            udf = fn.UserDataFunctions() # This is the instantiation of the class that is required to define User Data Functions

            @udf.function()
            def my_function() -> None:
                pass
    """

    def __init__(self):
        """ """
        self.logger = UdfLogger(__name__)
        try:
            self.__version__ = importlib.metadata.version("fabric_user_data_functions")
        except Exception as e:
            self.__version__ = "(version not found)"
        
        is_hosted_environment = os.environ.get('PYTHON_ENABLE_WORKER_EXTENSIONS') is not None
        if is_hosted_environment:
            self.logger.error(f"Fabric Python Worker Version: {self.__version__}")
        
        self._initialize_spec_generation()

        super().__init__(AuthLevel.ANONYMOUS)

    def function(self, name=None):
        """
        This decorator is used to define a User Data Function in Fabric. The function must be decorated with this decorator in order to be recognized as a User Data Function.

        :param name: The name of the function. This parameter is not used in the current version of Fabric.
        :type name: str

        .. remarks::

            .. code-block:: python
                import fabric.functions as fn

                udf = fn.UserDataFunctions()

                @udf.function() # This is the decorator that is required to define a User Data Function
                def my_function() -> None:
                    pass
        """
        if not self._spec_function_created:
            self._create_spec_function()

        def wrap(fb, user_func):
            # Add HTTP Trigger
            fb.add_trigger(
                trigger=HttpTrigger(
                    name=REQ_PARAMETER,
                    methods=[HttpMethod.POST],
                    auth_level=AuthLevel.ANONYMOUS,
                )
            )
            fb.add_binding(binding=HttpOutput(name="$return"))
            # Force one of our bindings to ensure the Host Extension is loaded
            fb.add_binding(
                binding=UserDataFunctionContextInput(
                    name=UNUSED_FABRIC_CONTEXT_PARAMETER
                )
            )
            # self.log_openapi_spec()
            return fb
        
        return fabric_decorators.configure_fabric_function_builder(self, wrap, self._functions_metadata)

    # The decorator that will be used to tell the function we want a fabric item
    def connection(
        self, alias: str, argName: Optional[str] = None, **kwargs
    ) -> Callable[..., Any]:
        """
        This decorator is used to tell a User Data Function that there is a connection to a data source. This decorator must be used in tandem with a parameter of type :class:`fabric.functions.FabricSqlConnection` or :class:`fabric.functions.FabricLakehouseClient` (see the example under `Remarks`).

        :param alias: The alias of the data connection that is being used.
        :type alias: str
        :param argName: The name of the parameter in the function signature. If not provided, the alias will be used.
        :type argName: str

        .. remarks::

            .. code-block:: python
                    import fabric.functions as fn

                    udf = fn.UserDataFunctions()

                    @udf.connection("<data connection alias>", "<argName>") # This is the decorator that is required to define a connection to a data source
                    @udf.function()
                    def my_function(<argName>: fn.FabricSqlConnection) -> None:
                        conn = <argName>.connect()
                        pass
        """

        @self._configure_function_builder
        def wrap(fb):

            fb.add_binding(
                binding=FabricItemInput(
                    name=argName if argName is not None else alias,
                    alias=alias,
                    **kwargs,
                )
            )
            return fb

        return wrap

    # The decorator that will be used to tell the function we want a fabric item
    def generic_connection(
        self, audienceType: str, argName: str, **kwargs
    ) -> Callable[..., Any]:
        """
        This decorator is used to tell a User Data Function that there is a connection to a data source. This decorator must be used in tandem with a parameter of type :class:`fabric.functions.FabricSqlConnection` or :class:`fabric.functions.FabricLakehouseClient` (see the example under `Remarks`).

        :param audienceType: The audience type of the data connection that is being used.
        :type audienceType: str
        :param argName: The name of the parameter in the function signature. If not provided, the audience type will be used.
        :type argName: str

        .. remarks::

            .. code-block:: python
                    import fabric.functions as fn

                    udf = fn.UserDataFunctions()

                    @udf.connection("<data connection audienceType>", "<argName>") # This is the decorator that is required to define a connection to a data source
                    @udf.function()
                    def my_function(<argName>: fn.FabricItem) -> None:
                        
                        pass
        """

        @self._configure_function_builder
        def wrap(fb):

            fb.add_binding(
                binding=FabricItemInput(
                    name=argName,
                    audienceType=audienceType,
                    **kwargs,
                )
            )
            return fb

        return wrap


    def context(self, argName, **kwargs) -> Callable[..., Any]:
        """
        This decorator is used to tell a User Data Function that there is a :class:`fabric.functions.UserDataFunctionContext` parameter.
        This decorator must be used in tandem with a parameter of type :class:`fabric.functions.UserDataFunctionContext`.

        :param argName: The name of the parameter in the function signature.
        :type argName: str

        .. remarks::

            .. code-block:: python
                    import fabric.functions as fn

                    udf = fn.UserDataFunctions()

                    @udf.context("<argName>") # This is the decorator that is required
                    @udf.function()
                    def my_function(<argName>: fn.UserDataFunctionContext) -> None:
                        pass
        """

        @self._configure_function_builder
        def wrap(fb):
            fb.add_binding(binding=UserDataFunctionContextInput(name=argName, **kwargs))
            return fb

        return wrap

    @deprecated(
        "This function is deprecated. Please use 'connection' instead. Note the alias parameter in 'connection' is now the first parameter."
    )
    def fabric_item_input(self, argName, alias: str, **kwargs) -> Callable[..., Any]:
        """
        This decorator is deprecated. Please use :meth:`connection` instead. Note the alias parameter in :meth:`connection` is now the first parameter.
        """
        return self.connection(alias, argName, **kwargs)

    @deprecated("This function is deprecated. Please use 'context' instead.")
    def user_data_function_context_input(self, argName, **kwargs) -> Callable[..., Any]:
        """
        This decorator is deprecated. Please use :meth:`context` instead.
        """
        return self.context(argName, **kwargs)
    
    def _create_spec_function(self):
        """
        This creates a function that is used to define the OpenAPI Specification for the User Data Functions.
        """
        self._spec_function_created = True
        @self.function()
        def get_udf_oai_spec_json(publicUrl: str, privateFunctions: list, specFormat: str = "yaml") -> str:
            """
            This function is used to generate the OpenAPI Specification for the User Data Functions in JSON format.

            :param publicUrl: The public URL of the User Data Functions.
            :type publicUrl: str
            :param privateFunctions: A list of private functions that should not be included in the OpenAPI Specification.
            :type privateFunctions: list
            :param specFormat: The format of the OpenAPI Specification. Defaults to 'yaml'.
            :type specFormat: str

            :return: The OpenAPI Specification for the User Data Functions in JSON format.
            :rtype: str
            """
            return self.get_udf_oai_spec(publicUrl, privateFunctions, specFormat)
        
        @self.function()
        def get_udf_oai_spec_yaml(publicUrl: str, privateFunctions: list, specFormat: str = "yaml") -> str:
            """
            This function is used to generate the OpenAPI Specification for the User Data Functions in YAML format.

            :param publicUrl: The public URL of the User Data Functions.
            :type publicUrl: str
            :param privateFunctions: A list of private functions that should not be included in the OpenAPI Specification.
            :type privateFunctions: list
            :param specFormat: The format of the OpenAPI Specification. Defaults to 'yaml'.
            :type specFormat: str

            :return: The OpenAPI Specification for the User Data Functions in YAML format.
            :rtype: str
            """
            return self.get_udf_oai_spec(publicUrl, privateFunctions, specFormat)

    def get_udf_oai_spec(self, publicUrl: str, privateFunctions: list, specFormat: str = "yaml") -> str:
        # Validation
        if specFormat not in ["json", "yaml"]:
            raise ValueError(
                f"Invalid specFormat: {specFormat}. Must be 'json' or 'yaml'."
            )
        
        if not publicUrl:
            publicUrl = SpecConstants.LOCAL_HOST_URL

        openapi_spec = OpenAPISpecification()
        _openapi_spec_builder = OpenAPISpecificationBuilder(openapi_spec)
        _openapi_spec_builder.with_openapi("3.0.0").with_info(
            title="Fabric User Data Functions",
            version="0.1.0",
            description="This is the OpenAPI Specification for Fabric User Data Functions.",
        ).add_security_scheme(
            name="bearerAuth",
            definition={
                "type": "http",
                "scheme": "bearer",
                "bearerFormat": "JWT",
            },
        ).add_security_requirement(
            {"bearerAuth": []}
        ).build()
        
        
        _functions_metadata_copy = copy.deepcopy(self._functions_metadata)
        
        # Remove private functions from the OpenAPI Spec
        for function in privateFunctions:
            self.logger.error(f"Probing function: {function}")
            for function_metadata in _functions_metadata_copy:
                self.logger.error(f"Function Metadata: {function_metadata.function_name}")
                if function_metadata.function_name == function:
                    self.logger.error(f"Removing function: {function}")
                    _functions_metadata_copy.remove(function_metadata)
        
        for function_metadata in _functions_metadata_copy:
            # For every key in the function_metadata path, add openapi spec path
            for key, value in function_metadata.paths.items():
                _openapi_spec_builder.add_path(key, value)
            for key, value in function_metadata.components.items():
                _openapi_spec_builder.add_schema_from_model(key, value)
    
        _openapi_spec_builder.add_server(publicUrl, SpecConstants.PUBLIC_URL_DESC).build()
        self.logger.error(f"OpenAPI Spec: {openapi_spec.serialize(specFormat)}")
        return openapi_spec.serialize(specFormat)
    
    def _initialize_spec_generation(self):
        self._spec_function_created = False
        self._functions_metadata: list[FunctionMetadata] = []
        



# flake8: noqa: I003

import importlib

class module_factory:
    
    modules = {}

    def __init__(self):
        pass

    def load_module(self, module_name:str):
        if module_name in module_factory.modules:
            return module_factory.modules[module_name]
        else:
            try:
                module_factory.modules[module_name] = importlib.import_module(module_name)
                return module_factory.modules[module_name]
            except ImportError as e:
                raise ImportError(f"Module {module_name} not found") from e


from typing import Optional


def extract_summary_description(docstring: str) -> tuple[Optional[str], Optional[str]]:
    """
    Retrieves the Summary and Description from a multiline comment string.

    Args:
        docstring: A multiline string containing Summary and Description.

    Returns:
        A tuple containing two strings: (summary, description).
        If Summary or Description is not found, the corresponding value will be None.
    """
    summary_lines = []
    description_lines = []
    in_summary = False
    in_description = False

    # Proceed if docstring is not None
    if docstring is None:
        return None, None
    for line in docstring.splitlines():
        stripped_line = line.strip()

        if stripped_line.startswith("Summary:"):
            in_summary = True
            in_description = False
            summary_lines.append(stripped_line[len("Summary:"):].strip())
        elif stripped_line.startswith("Description:"):
            in_summary = False
            in_description = True
            description_lines.append(stripped_line[len("Description:"):].strip())
        elif in_summary:
            summary_lines.append(stripped_line)
        elif in_description:
            description_lines.append(stripped_line)

    summary = None
    if summary_lines:
        summary = "\n".join(summary_lines).strip()
    description = None
    if description_lines:
        description = "\n".join(description_lines).strip()
    return summary, description



# flake8: noqa: TAE002
from pydantic import BaseModel, create_model
from typing import Dict, Any, List, Optional, Type, Union
import inspect

CellValue = Union[
    str,
    float,
    int,
    bool,
    Dict[str, Any],
    List[Any],
]

ColumnData = List[CellValue]

class FunctionMetadata:
    def __init__(self, name: str):
        self.function_name = name
        self.paths = {}
        self.components = {}

    def add_path(self, path_name: str, path_details: dict):
        self.paths[path_name] = path_details
    
    def add_component(self, component_name: str, component: dict):
        self.components[component_name] = component


class OpenAPISpecification:
    """
    Represents the OpenAPI specification object.
    """

    def __init__(self):
        self.openapi: Optional[str] = None
        self.info: Optional[Dict[str, Any]] = None
        self.servers: Optional[List[Dict[str, Any]]] = None
        self.paths: Optional[Dict[str, Dict[str, Any]]] = None
        self.components: Optional[Dict[str, Dict[str, Any]]] = None
        self.security: Optional[List[Dict[str, List[str]]]] = None
        self.tags: Optional[List[Dict[str, Any]]] = None
        self.externalDocs: Optional[Dict[str, str]] = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the OpenAPI specification object to a Python dictionary.
        """
        spec = {}
        if self.openapi:
            spec["openapi"] = self.openapi
        if self.info:
            spec["info"] = self.info
        if self.servers:
            spec["servers"] = self.servers
        if self.paths:
            spec["paths"] = self.paths
        if self.components:
            spec["components"] = self.components
        if self.security:
            spec["security"] = self.security
        if self.tags:
            spec["tags"] = self.tags
        if self.externalDocs:
            spec["externalDocs"] = self.externalDocs
        return spec
    
    def to_json(self) -> str:
        """
        Converts the OpenAPI specification object to a JSON string.
        """
        import json
        return json.dumps(self.to_dict(), indent=2)
    
    def to_yaml(self) -> str:
        """
        Converts the OpenAPI specification object to a YAML string.
        """
        import yaml
        return yaml.dump(self.to_dict(), indent=2)
    
    def serialize(self, format: str = "json") -> str:
        """
        Serializes the OpenAPI specification object to a string in the specified format.
        
        Args:
            format: The serialization format to use. Can be "json" or "yaml".
        """
        if format == "json":
            return self.to_json()
        elif format == "yaml":
            return self.to_yaml()
        else:
            raise ValueError("Invalid format. Must be 'json' or 'yaml'.")


class OpenAPISpecificationBuilder:
    """
    Builder class for constructing the OpenAPI specification object step by step.
    """

    def __init__(self, spec: Optional[OpenAPISpecification]):
        self._spec = spec

    def with_openapi(self, version: str) -> "OpenAPISpecificationBuilder":
        """
        Sets the OpenAPI version.
        """
        self._spec.openapi = version
        return self

    def with_info(
        self, title: str, version: str, description: Optional[str] = None
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the info section of the OpenAPI specification.
        """
        self._spec.info = {
            "title": title,
            "version": version,
        }
        if description:
            self._spec.info["description"] = description
        return self

    def with_servers(
        self, servers: List[Dict[str, Any]]
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the servers section of the OpenAPI specification.

        Args:
            servers: A list of server objects (dictionaries) as defined in the OpenAPI specification.
        """
        self._spec.servers = servers
        return self

    def add_server(
        self,
        url: str,
        description: Optional[str] = None,
        variables: Optional[Dict[str, Any]] = None,
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a single server to the servers section of the OpenAPI specification. If the URL is already present, it will not be added again.

        Args:
            url: The URL of the server.
            description: An optional description of the server.
            variables: An optional dictionary of variables used in the URL template.
        """
        # Append only if url is unique
        if self._spec.servers is None:
            self._spec.servers = []
        if not any(server["url"] == url for server in self._spec.servers):
            server = {"url": url}
            if description:
                server["description"] = description
            if variables:
                server["variables"] = variables
            self._spec.servers.append(server)
        return self

    def with_paths(
        self, paths: Dict[str, Dict[str, Any]]
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the paths section of the OpenAPI specification.
        """
        self._spec.paths = paths
        return self

    def add_path(
        self,
        path: str,
        operations: Dict[str, Any],
        operation_ids: Optional[Dict[str, str]] = None,
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a single path and its operations to the specification.

        Args:
            path: The path URL (e.g., "/items").
            operations: A dictionary where keys are HTTP methods (e.g., "get", "post")
                        and values are the corresponding operation definitions.
            operation_ids: An optional dictionary where keys are HTTP methods and values
                           are the unique operationIds for those methods.
        """
        if self._spec.paths is None:
            self._spec.paths = {}
        self._spec.paths[path] = operations
        if operation_ids:
            for method, operation_id in operation_ids.items():
                if method in self._spec.paths[path] and operation_id:
                    self._spec.paths[path][method]["operationId"] = operation_id
        return self

    def with_components(
        self, components: Dict[str, Dict[str, Any]]
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the components section of the OpenAPI specification.
        """
        self._spec.components = components
        return self

    def add_component(
        self, component_type: str, name: str, definition: Dict[str, Any]
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a single component to the specification.
        component_type can be 'schemas', 'responses', 'parameters', etc.
        """
        if self._spec.components is None:
            self._spec.components = {}
        if component_type not in self._spec.components:
            self._spec.components[component_type] = {}
        self._spec.components[component_type][name] = definition
        return self

    def add_schema_from_model(
        self, schema_name: str, model_class: Type[BaseModel]
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a schema definition to the components section based on a Pydantic model.

        Args:
            schema_name: The name to use for the schema in the components section.
            model_class: The Pydantic model class to generate the schema from.
        """
        schema = model_class.model_json_schema()
        self.add_component("schemas", schema_name, schema)
        return self

    def add_security_scheme(
        self, name: str, definition: Dict[str, Any]
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a security scheme definition to the components section.

        Args:
            name: The name to use for the security scheme.
            definition: A dictionary defining the security scheme (e.g., type, description, etc.).
        """
        if self._spec.components is None:
            self._spec.components = {}
        if "securitySchemes" not in self._spec.components:
            self._spec.components["securitySchemes"] = {}
        self._spec.components["securitySchemes"][name] = definition
        return self

    def with_security(
        self, security: List[Dict[str, List[str]]]
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the security section at the root level of the OpenAPI specification.
        This defines global security requirements.
        """
        self._spec.security = security
        return self

    def add_security_requirement(
        self, security_requirement: Dict[str, List[str]]
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a security requirement to the root level security section.
        This defines a global security requirement that applies to all operations.
        """
        if self._spec.security is None:
            self._spec.security = []
        self._spec.security.append(security_requirement)
        return self

    def with_tags(self, tags: List[Dict[str, Any]]) -> "OpenAPISpecificationBuilder":
        """
        Sets the tags section of the OpenAPI specification.
        """
        self._spec.tags = tags
        return self

    def add_tag(
        self, name: str, description: Optional[str] = None
    ) -> "OpenAPISpecificationBuilder":
        """
        Adds a single tag to the specification.
        """
        if self._spec.tags is None:
            self._spec.tags = []
        tag = {"name": name}
        if description:
            tag["description"] = description
        self._spec.tags.append(tag)
        return self

    def with_external_docs(
        self, description: str, url: str
    ) -> "OpenAPISpecificationBuilder":
        """
        Sets the externalDocs section of the OpenAPI specification.
        """
        self._spec.externalDocs = {
            "description": description,
            "url": url,
        }
        return self

    def build(self) -> OpenAPISpecification:
        """
        Builds and returns the final OpenAPI specification object.
        """
        return self._spec


def create_pydantic_model(
    model_name: str, parameters: List[inspect.Parameter]
) -> Type[BaseModel]:
    """
    Dynamically creates a Pydantic model class from a list of inspect.Parameter
    objects, converting pandas DataFrame/Series annotations to compatible types.

    Args:
        model_name: The name of the Pydantic model to create.
        parameters: A list of inspect.Parameter objects representing the model's fields.

    Returns:
        The dynamically created Pydantic model class.

    Raises:
        ImportError: If pandas is not installed but DataFrame/Series annotations are used.
    """
    processed_params_dict = {}  # Use a dict to build field definitions

    for param in parameters:
        annotation = param.annotation
        param_name = param.name
        param_default = param.default

        final_annotation = Any  # Default if no annotation
        final_default = ...  # Default to required (Ellipsis)

        # 1. Determine the correct annotation
        if annotation is not inspect.Parameter.empty:
            try:
                # Check for DataFrame
                if annotation.__name__ == "DataFrame":
                    final_annotation = Dict[str, ColumnData]
                # Check for Series
                elif annotation.__name__ == "Series":
                    final_annotation = Dict[str, ColumnData]
                # Otherwise, use the original annotation
                else:
                    final_annotation = annotation
            except Exception:  # Catch other potential issues with the annotation type
                final_annotation = Any
        else:
            # No annotation provided
            final_annotation = Any

        # 2. Determine the default value (or required status)
        if param_default is not inspect.Parameter.empty:
            final_default = param_default  # Use the actual default value
        # else: keep final_default = ... (Ellipsis for required)

        # 3. Add to field definitions dictionary
        processed_params_dict[param_name] = (final_annotation, final_default)

    # 4. Create the Pydantic model using the processed field definitions
    try:
        # Use type('BaseModel', (BaseModel,), {}) as base if BaseModel directly causes issues
        # But usually inheriting from BaseModel is fine.
        return create_model(
            model_name, __base__=BaseModel, **processed_params_dict
        )
    except Exception as e:
        error_msg = f"Error creating Pydantic model '{model_name}': {str(e)}"
        raise Exception(error_msg)



# flake8: noqa: I900
class GenericAudienceTypes:
    """
    This class contains the different types of generic audiences.
    .. remarks::

        .. note::
            This class is used to define the types of generic audiences that can be used in the generic_connection decorator.
    """
    CosmosDB = "CosmosDB"
    KeyVault = "KeyVault"


# flake8: noqa: I005
import pyarrow as pa

class arrow_dataframe_response():

    def __init__(self, df):
        self._df = df

    def to_bytes(self) -> bytes:
        sink = pa.BufferOutputStream()
        batch_arrow = pa.RecordBatch.from_pandas(self._df)
        writer = pa.ipc.new_stream(sink, batch_arrow.schema)
        writer.write_batch(batch_arrow)
        writer.close()
        return sink.getvalue().to_pybytes()



# flake8: noqa: I900, R504
from urllib.parse import urlparse

from azure.storage.filedatalake import DataLakeDirectoryClient
from azure.storage.filedatalake.aio import DataLakeDirectoryClient as DataLakeDirectoryClientAsync

from fabric.internal.custom_token_credential import \
    CustomTokenCredential
from fabric.functions.fabric_item import FabricItem
from fabric.functions.udf_exception import UserDataFunctionInternalError


class FabricLakehouseFilesClient(FabricItem):
    """
    This class is used to connect to resources that supply Lakehouse file connection strings.

    .. remarks::

        .. note::
            If you want to connect to a Lakehouse Files endpoint, that should be done through
            :class:`FabricLakehouseClient` instead (and then use it's method :meth:`FabricLakehouseClient.connectToFiles`).

    """

    __APPSETTINGS_PATH = "fileendpoint"

    def connect(self) -> DataLakeDirectoryClient:
        """Makes a connection to the Lakehouse Files endpoint. Call this within your function.
        """
        if self.__APPSETTINGS_PATH not in self.endpoints:
            raise UserDataFunctionInternalError(f"{self.__APPSETTINGS_PATH} is not set")

        raw_path = self.endpoints[self.__APPSETTINGS_PATH]['ConnectionString']
        parsed_path = urlparse(raw_path)

        access_token = self.endpoints[self.__APPSETTINGS_PATH]['AccessToken']

        # The account URL is the scheme and netloc parts of the parsed path
        account_url = f"{parsed_path.scheme}://{parsed_path.netloc}"

        # The file system name and directory name are in the path part of the parsed path
        # We remove the leading slash and then split the rest into the file system name
        # and directory name
        file_system_name, _, directory_name = parsed_path.path.lstrip('/').partition('/')

        directory_client = DataLakeDirectoryClient(account_url, file_system_name, directory_name,
                                                    CustomTokenCredential(access_token))
        return directory_client

    def connect_async(self) -> DataLakeDirectoryClientAsync:
        """Makes a connection to the Lakehouse Files endpoint. Call this within your function. Returns a client that can be used in async functions.
        """
        if self.__APPSETTINGS_PATH not in self.endpoints:
            raise UserDataFunctionInternalError(f"{self.__APPSETTINGS_PATH} is not set")

        raw_path = self.endpoints[self.__APPSETTINGS_PATH]['ConnectionString']
        parsed_path = urlparse(raw_path)

        access_token = self.endpoints[self.__APPSETTINGS_PATH]['AccessToken']

        # The account URL is the scheme and netloc parts of the parsed path
        account_url = f"{parsed_path.scheme}://{parsed_path.netloc}"

        # The file system name and directory name are in the path part of the parsed path
        # We remove the leading slash and then split the rest into the file system name
        # and directory name
        file_system_name, _, directory_name = parsed_path.path.lstrip('/').partition('/')

        directory_client = DataLakeDirectoryClientAsync(account_url, file_system_name, directory_name,
                                                    CustomTokenCredential(access_token))
        return directory_client



# """You can ignore this module and any files within it. It is used to help set up your User Data Functions.
# """

# flake8: noqa: I005
from .arrow_request import arrow_request
from .arrow_dataframe_response import arrow_dataframe_response
from .arrow_series_response import arrow_series_response



import importlib
import json
import logging
import os
from pathlib import Path
import sys
import subprocess
# flake8: noqa: I900
import azure.functions as azure_func

# return code:
# 0 - no error
# 1 - error

PYTHON_VERSION_REQUIRED = (3, 11)
SCRIPT_FILE_NAME = 'function_app.py'
METADATA_FILE_NAME = 'functions.metadata'
FABRIC_UDF_LIB_NAME = 'fabric_user_data_functions'
NEEDED_BINDING_TYPES = {'httpTrigger', 'UdfProperty', 'FabricItem', 'UserDataFunctionContext'}
SUPPORTED_UDF_PARAMETER_TYPES = ["str", "int", "float", "bool", "datetime", "list", "dict", "set", "tuple"]
SUPPORTED_UDF_RETURN_TYPES = list(SUPPORTED_UDF_PARAMETER_TYPES) + ["None"]
ERROR_CODE = 1
SUCCESS = 0

def _get_app_functions(file_name: str = SCRIPT_FILE_NAME):
    module_path = os.getcwd()
    func_file = os.path.join(module_path, file_name)

    if not os.path.exists(func_file):
        logging.error(f'Cannot find source code {file_name}')
        return None
    
    sys.path.append(str(module_path))
    module_name = Path(func_file).stem
    imported_module = importlib.import_module(module_name)

    app = None
    for i in imported_module.__dir__():
        if isinstance(getattr(imported_module, i, None), azure_func.FunctionRegister):
            if not app:
                app = getattr(imported_module, i, None)
            else:
                logging.error("More than 1 app found in the source code")
                return None

    if not app:
        logging.error("No app found in the source code")
        return None
    
    functions = app.get_functions()
    if not functions:
        logging.error("No function found in the source code")
        return None
    
    return functions

def _collect_function_metadata(functions):
    metadata=[]
    warnings = []
    errors = []
    for func in functions:
        # log the function bindings
        logging.info(f'Getting metadata from Function {func._name}')

        single_func_meta = {}
        single_func_meta['name'] = func._name
        single_func_meta['scriptFile'] = func.function_script_file
        single_func_meta['bindings'] = []

        udf_worker_version = _get_fabric_lib_version()
        if udf_worker_version is not None and len(udf_worker_version) > 0:
            single_func_meta['fabricWorkerVersion'] = udf_worker_version

        for binding in func._bindings:
            if binding.type not in NEEDED_BINDING_TYPES:
                continue
            logging.info(f' Binding name: {binding.name}')
            logging.info(f' Binding type: {binding.type}')
            logging.info(f' Binding direction: {binding._direction.name}')

            b = {}
            b['name'] = binding.name
            b['direction'] = ''.join([binding._direction.name[0], binding._direction.name[1:].lower()])
            b['type'] = binding.type

            additional_props = {}
            if binding.type == 'httpTrigger':
                additional_props = _get_http_binding_properties(binding)

            if binding.type == 'UdfProperty':
                additional_props = _get_udf_binding_properties(binding)
            
            if binding.type == 'FabricItem':
                additional_props = _get_fabricItem_binding_properties(binding)

            if binding.type == 'UserDataFunctionContext':
                additional_props = _get_userDataFunctionContext_binding_properties(binding)

            for key, value in additional_props.items():
                b[key] = value

            single_func_meta['bindings'].append(b)

        udf_properties, udfproperty_errors = _get_udf_properties_from_function(single_func_meta, func)
        if (len(udfproperty_errors) > 0):
            errors.extend(udfproperty_errors)

        single_func_meta['fabricProperties'] = udf_properties

        # remove UdfProperties from bindings
        single_func_meta['bindings'] = [b for b in single_func_meta['bindings'] if b['type'] != 'UdfProperty']

        metadata.append(single_func_meta)
    
    warning = _generate_metadata_warnings(metadata)
    if(warning is not None and len(metadata) > 0):
        warnings.append(warning)
    
    # put all warnings/errors in the first function because metadata is an array of functions, and the vscode extension knows to look here for them

    if len(warnings) > 0:
        metadata[0]['warnings'] = warnings
    
    if len(errors) > 0:
        metadata[0]['errors'] = errors
    
    return metadata


def _get_fabric_lib_version():
    try:
        output = subprocess.check_output([sys.executable, '-m', 'pip', 'show', FABRIC_UDF_LIB_NAME], text=True)
        version_line = next((line for line in output.splitlines() if line.startswith('Version:')), None)
        if version_line:
            return version_line.split(':')[1].strip()
    except subprocess.CalledProcessError:
        logging.error(f'Cannot find the version of {FABRIC_UDF_LIB_NAME}')
        return None

def _get_http_binding_properties(binding):
    props = {}
    methods = getattr(binding, 'methods', None)
    if not methods:
        methods = [azure_func.HttpMethod.POST]
    
    props['methods'] = []
    for m in methods:
        if m == azure_func.HttpMethod.PUT:
            logging.error("Fabric Function currently does not support PUT HTTP actions.")
            return None
        props['methods'].append(m.value.lower())

    props['route'] = getattr(binding, 'route', None)
    auth = getattr(binding, 'auth_level', None)
    if auth:
        props['authLevel'] = ''.join([auth.value[0].upper(), auth.value[1:].lower()])

    return props

def _get_fabricItem_binding_properties(binding):
    props = {}
    
    props['alias'] = getattr(binding, 'alias', None)

    return props

def _get_userDataFunctionContext_binding_properties(binding):
    props = {}
    
    props['parameterName'] = getattr(binding, 'parameterName', None)

    return props

def _get_udf_binding_properties(binding):
    props = {}
    
    props['parameterName'] = getattr(binding, 'parameterName', None)
    typeName = getattr(binding, 'typeName', 'any')
    if (typeName == 'UdfString'):
        typeName = 'str'
    props['typeName'] = typeName

    return props

   
def _ensure_return_type_is_supported(return_type_name: str, func_name: str, errors: list):
    if return_type_name not in SUPPORTED_UDF_RETURN_TYPES:
        errors.append(f"The function '{func_name}' has a return type '{return_type_name}' that is not supported. Supported types are: {SUPPORTED_UDF_RETURN_TYPES}")

def _ensure_parameter_type_is_supported(parameter_name: str, parameter_type_name: str, func_name: str, errors: list):
     # If the parameter has a fabricitem binding, this function won't check its argument type, but the error message should probably contain that those are available as data types to avoid user confusion.
    if parameter_type_name not in SUPPORTED_UDF_PARAMETER_TYPES:
        errors.append(f"The function '{func_name}' has a parameter '{parameter_name}' with a data type '{parameter_type_name}' that is not supported. Supported types are: {SUPPORTED_UDF_PARAMETER_TYPES}")

def _get_udf_properties_from_function(func_meta, func):
    errors = []
    props = {}

    if 'old_return' not in func._func.__annotations__:
        errors.append(f"Function '{func_meta['name']}': return type must be annotated.")
    else:
        return_type = getattr(func._func.__annotations__['old_return'], "__name__", "None")
        _ensure_return_type_is_supported(return_type, func_meta['name'], errors)
        props['fabricFunctionReturnType'] = return_type
    
    props['fabricFunctionParameters'] = []

    for binding in func_meta['bindings']:
        if binding['type'] == 'UdfProperty':
            p = {}
            p['name'] = binding['parameterName']
            p['dataType'] = binding['typeName']
            _ensure_parameter_type_is_supported(p['name'], p['dataType'], func_meta['name'], errors)
            props['fabricFunctionParameters'].append(p)

    return props, errors

def _get_metadata_file_path(file_name=METADATA_FILE_NAME):
    base_directory = os.getcwd()
    return os.path.join(base_directory, "fabric_lib", file_name)

def _write_metadata_file(metadata, file_name=METADATA_FILE_NAME):
    metadata_json = json.dumps(metadata)
    output = _get_metadata_file_path(file_name)
    with open(output, 'w') as meta_file:
        meta_file.write(metadata_json)

def _delete_metadata_file_if_exist(file_name=METADATA_FILE_NAME):
    metadata_file = _get_metadata_file_path(file_name)
    if os.path.exists(metadata_file):
        os.remove(metadata_file)

def _get_sys_version():
    return sys.version_info

def _generate_metadata_warnings(metadata):
    current_version = _get_sys_version()
    if current_version[0] != PYTHON_VERSION_REQUIRED[0] or current_version[1] != PYTHON_VERSION_REQUIRED[1]:
        return f'Python version {current_version[0]}.{current_version[1]} is not equal to the required version {PYTHON_VERSION_REQUIRED[0]}.{PYTHON_VERSION_REQUIRED[1]}'

def generate_function_metadata():
    try:
        # remove metadata first since VSCode extension uses existence of the file as the success of the script run
        _delete_metadata_file_if_exist()

        functions = _get_app_functions()

        if(functions is None):
            return ERROR_CODE
        
        metadata = _collect_function_metadata(functions)
        if(metadata is None):
            return ERROR_CODE

        _write_metadata_file(metadata)
        # prettify the json metadata and log it
        logging.info(json.dumps(metadata, indent=2, sort_keys=True))
        logging.info(f'Function metadata generated successfully for {len(functions)} functions.')
        return SUCCESS
    except Exception as e:
        logging.error(f'Got errors when generating metadata. Details:{e}')
        return ERROR_CODE
